@@ -239,6 +239,7 @@ typedef struct {
 	chunkqueue in_queue;
 } handler_ctx;
 
+__attribute_returns_nonnull__
 static handler_ctx *handler_ctx_init(void) {
 	handler_ctx * const hctx = ck_calloc(1, sizeof(*hctx));
 	chunkqueue_init(&hctx->in_queue);
@@ -1517,16 +1518,24 @@ static int mod_deflate_using_libdeflate (handler_ctx * const hctx, const plugin_
     buffer * const fn = hctx->output; /*(&p->tmp_buf)*/
     int fd = hctx->cache_fd;
     if (-1 == fd) {
+        /* create temp file in temp chunkqueue and pluck from chunkqueue */
+        #if 0
+        chunkqueue tq = {0,0,0,0,0,0}; /*(fake cq for tempfile creation)*/
+        chunkqueue_init(&tq);
+        #else
         chunkqueue * const cq = &hctx->r->write_queue;
-        if (cq->tempdirs && cq->tempdir_idx < cq->tempdirs->used) {
-            data_string *ds =(data_string *)cq->tempdirs->data[cq->tempdir_idx];
-            buffer_copy_string_len(fn, BUF_PTR_LEN(&ds->value));
-        }
-        else
-            buffer_copy_string_len(fn, CONST_STR_LEN("/var/tmp"));
-        buffer_append_path_len(fn, CONST_STR_LEN("lighttpd-XXXXXX"));
-        fd = fdevent_mkostemp(fn->ptr, 0);
-        if (-1 == fd) return 0;
+        chunkqueue tq = *cq; /* copy struct, including tempdir state */
+        tq.first = tq.last = NULL; /* discard duplicated chunks from orig cq */
+        #endif
+        if (0 != chunkqueue_append_mem_to_tempfile(&tq,"",0,hctx->r->conf.errh))
+            return 0;
+        /* copy temp file fd and fn from temp chunkqueue tq and then reset tq */
+        chunk * const c = tq.last;
+        fd = c->file.fd;
+        c->file.fd = -1;
+        buffer_copy_buffer(fn, c->mem);
+        buffer_clear(c->mem);
+        chunkqueue_reset(&tq);
     }
 
     const size_t sz =
@@ -1893,16 +1902,17 @@ REQUEST_FUNC(mod_deflate_handle_response_start) {
 		return HANDLER_GO_ON;
 
 	/* disable compression for some http status types. */
+	if (r->http_status < 200)
+		return HANDLER_GO_ON; /* r->http_status is 1xx intermed response */
 	switch(r->http_status) {
-	case 100:
-	case 101:
+	case 200: /* common case */
+	default:
+		break;
 	case 204:
 	case 205:
 	case 304:
 		/* disable compression as we have no response entity */
 		return HANDLER_GO_ON;
-	default:
-		break;
 	}
 
 	mod_deflate_patch_config(r, p);
