@@ -51,13 +51,22 @@ static stat_cache sc;
 
 
 __attribute_noinline__
+static void * stat_cache_sptree_ndx(splay_tree ** const sptree,
+                                    int * const ndxp,
+                                    const char * const name,
+                                    uint32_t len)
+{
+    const int ndx = splaytree_djbhash(name, len);
+    if (ndxp) *ndxp = ndx;
+    *sptree = splaytree_splay(*sptree, ndx);
+    return (*sptree && (*sptree)->key == ndx) ? (*sptree)->data : NULL;
+}
+
 static void * stat_cache_sptree_find(splay_tree ** const sptree,
                                      const char * const name,
                                      uint32_t len)
 {
-    const int ndx = splaytree_djbhash(name, len);
-    *sptree = splaytree_splay(*sptree, ndx);
-    return (*sptree && (*sptree)->key == ndx) ? (*sptree)->data : NULL;
+    return stat_cache_sptree_ndx(sptree, NULL, name, len);
 }
 
 
@@ -300,11 +309,10 @@ static void fam_dir_periodic_cleanup(void) {
         max_ndx = 0;
         fam_dir_tag_refcnt(scf->dirs, keys, &max_ndx);
         for (i = 0; i < max_ndx; ++i) {
-            const int ndx = keys[i];
-            splay_tree *node = scf->dirs = splaytree_splay(scf->dirs, ndx);
-            if (node && node->key == ndx) {
-                fam_dir_entry *fam_dir = node->data;
-                scf->dirs = splaytree_delete(scf->dirs, ndx);
+            {
+                scf->dirs = splaytree_splay_nonnull(scf->dirs, keys[i]);
+                fam_dir_entry *fam_dir = scf->dirs->data;
+                scf->dirs = splaytree_delete_splayed_node(scf->dirs);
               #ifdef HAVE_SYS_INOTIFY_H
                 scf->wds = splaytree_delete(scf->wds, fam_dir->req);
               #elif defined HAVE_SYS_EVENT_H && defined HAVE_KQUEUE
@@ -630,8 +638,7 @@ static void stat_cache_free_fam(stat_cache_fam *scf) {
 
       #ifdef HAVE_SYS_INOTIFY_H
 	while (scf->wds) {
-		splay_tree *node = scf->wds;
-		scf->wds = splaytree_delete(scf->wds, node->key);
+		scf->wds = splaytree_delete_splayed_node(scf->wds);
 	}
       #elif defined HAVE_SYS_EVENT_H && defined HAVE_KQUEUE
 	/*(quicker cleanup to close kqueue() before cancel per entry)*/
@@ -640,9 +647,8 @@ static void stat_cache_free_fam(stat_cache_fam *scf) {
       #endif
 	while (scf->dirs) {
 		/*(skip entry invalidation and FAMCancelMonitor())*/
-		splay_tree *node = scf->dirs;
-		fam_dir_entry_free((fam_dir_entry *)node->data);
-		scf->dirs = splaytree_delete(scf->dirs, node->key);
+		fam_dir_entry_free((fam_dir_entry *)scf->dirs->data);
+		scf->dirs = splaytree_delete_splayed_node(scf->dirs);
 	}
 
 	if (-1 != scf->fd) {
@@ -670,12 +676,11 @@ static fam_dir_entry * fam_dir_monitor(stat_cache_fam *scf, char *fn, uint32_t d
         while (fn[--dirlen] != '/') ;
         if (0 == dirlen) dirlen = 1; /*(should not happen for file)*/
     }
-    int dir_ndx = splaytree_djbhash(fn, dirlen);
-    fam_dir_entry *fam_dir = NULL;
+    int dir_ndx;
+    fam_dir_entry *fam_dir =
+      stat_cache_sptree_ndx(&scf->dirs, &dir_ndx, fn, dirlen);
 
-    scf->dirs = splaytree_splay(scf->dirs, dir_ndx);
-    if (NULL != scf->dirs && scf->dirs->key == dir_ndx) {
-        fam_dir = scf->dirs->data;
+    if (NULL != fam_dir) {
         if (!buffer_eq_slen(&fam_dir->name, fn, dirlen)) {
             /* hash collision; preserve existing
              * do not monitor new to avoid cache thrashing */
@@ -731,7 +736,7 @@ static fam_dir_entry * fam_dir_monitor(stat_cache_fam *scf, char *fn, uint32_t d
             fam_dir->st_dev = st->st_dev;
             fam_dir->st_ino = st->st_ino;
           #ifdef HAVE_SYS_INOTIFY_H
-            scf->wds = splaytree_insert(scf->wds, fam_dir->req, fam_dir);
+            scf->wds = splaytree_insert_splayed(scf->wds,fam_dir->req,fam_dir);
           #endif
         }
         fam_dir->stat_ts = cur_ts;
@@ -756,7 +761,7 @@ static fam_dir_entry * fam_dir_monitor(stat_cache_fam *scf, char *fn, uint32_t d
             return NULL;
         }
 
-        scf->dirs = splaytree_insert(scf->dirs, dir_ndx, fam_dir);
+        scf->dirs = splaytree_insert_splayed(scf->dirs, dir_ndx, fam_dir);
       #ifdef HAVE_SYS_INOTIFY_H
         scf->wds = splaytree_insert(scf->wds, fam_dir->req, fam_dir);
       #endif
@@ -831,6 +836,12 @@ void stat_cache_entry_refchg(void *data, int mod) {
         sce->refcnt += mod;
 }
 
+__attribute_nonnull__()
+static splay_tree * stat_cache_sptree_node_free(splay_tree *sptree) {
+    stat_cache_entry_free(sptree->data);
+    return splaytree_delete_splayed_node(sptree);
+}
+
 #if defined(HAVE_XATTR) || defined(HAVE_EXTATTR)
 
 static const char *attrname = "Content-Type";
@@ -841,8 +852,13 @@ static int stat_cache_attr_get(const char *name) {
   #if defined(HAVE_XATTR)
    #if defined(HAVE_SYS_XATTR_H)
     ssize_t attrlen;
+   #if defined(__APPLE__) && defined(__MACH__)
+    if (0 < (attrlen = getxattr(name, attrname,
+                                attrval, sizeof(attrval)-1, 0, 0)))
+   #else
     if (0 < (attrlen = getxattr(name, attrname,
                                 attrval, sizeof(attrval)-1)))
+   #endif
    #else
     int attrlen = sizeof(attrval)-1;
     if (0 == attr_get(name, attrname, attrval, &attrlen, 0))
@@ -879,8 +895,7 @@ int stat_cache_init(fdevents *ev, log_error_st *errh) {
 void stat_cache_free(void) {
     splay_tree *sptree = sc.files;
     while (sptree) {
-        stat_cache_entry_free(sptree->data);
-        sptree = splaytree_delete(sptree, sptree->key);
+        sptree = stat_cache_sptree_node_free(sptree);
     }
     sc.files = NULL;
 
@@ -1133,8 +1148,7 @@ void stat_cache_delete_entry(const char *name, uint32_t len)
     splay_tree **sptree = &sc.files;
     stat_cache_entry *sce = stat_cache_sptree_find(sptree, name, len);
     if (sce && buffer_is_equal_string(&sce->name, name, len)) {
-        stat_cache_entry_free(sce);
-        *sptree = splaytree_delete(*sptree, (*sptree)->key);
+        *sptree = stat_cache_sptree_node_free(*sptree);
     }
 }
 
@@ -1211,12 +1225,8 @@ static void stat_cache_prune_dir_tree(const char *name, size_t len)
         max_ndx = 0;
         stat_cache_tag_dir_tree(sptree, name, len, keys, &max_ndx);
         for (i = 0; i < max_ndx; ++i) {
-            const int ndx = keys[i];
-            splay_tree *node = sptree = splaytree_splay(sptree, ndx);
-            if (node && node->key == ndx) {
-                stat_cache_entry_free(node->data);
-                sptree = splaytree_delete(sptree, ndx);
-            }
+            sptree = splaytree_splay_nonnull(sptree, keys[i]);
+            sptree = stat_cache_sptree_node_free(sptree);
         }
     } while (max_ndx == sizeof(keys)/sizeof(int));
     sc.files = sptree;
@@ -1245,147 +1255,128 @@ void stat_cache_delete_dir(const char *name, uint32_t len)
   #endif
 }
 
-/***
- *
- *
- *
- * returns:
- *  - HANDLER_FINISHED on cache-miss (don't forget to reopen the file)
- *  - HANDLER_ERROR on stat() failed -> see errno for problem
- */
-
-stat_cache_entry * stat_cache_get_entry(const buffer * const name) {
-	stat_cache_entry *sce = NULL;
-
-	/* consistency: ensure lookup name does not end in '/' unless root "/"
-	 * (but use full path given with stat(), even with trailing '/') */
-	int final_slash = 0;
-	size_t len = buffer_clen(name);
-	force_assert(0 != len);
-	if (name->ptr[len-1] == '/') { final_slash = 1; if (0 == --len) len = 1; }
-	/* Note: paths are expected to be normalized before calling stat_cache,
-	 * e.g. without repeated '/' */
+__attribute_cold__
+__attribute_noinline__
+static stat_cache_entry * stat_cache_refresh_entry(const buffer * const name, uint32_t len, stat_cache_entry *sce, const int file_ndx, const int refresh) {
 
   #ifndef _WIN32
-	if (name->ptr[0] != '/') {
-		errno = EINVAL;
-		return NULL;
-	}
+    /* sanity check; should not happen; should not be called with rel paths */
+    if (__builtin_expect( (name->ptr[0] != '/'), 0)) {
+        errno = EINVAL;
+        return NULL;
+    }
   #endif
 
-	/*
-	 * check if the directory for this file has changed
-	 */
+    /* use full path w/ stat(), even w/ trailing '/' ('len' may be shorter) */
+    struct stat st;
+    if (-1 == stat(name->ptr, &st))
+        return NULL;
 
-	const unix_time64_t cur_ts = log_monotonic_secs;
+    if (NULL == sce || !stat_cache_stat_eq(&sce->st, &st)) {
+        if (NULL != sce && sce->fd >= 0) {
+            /* close fd when refresh needed */
+            if (1 == sce->refcnt) {
+                close(sce->fd);
+                sce->fd = -1;
+            }
+            else {
+                --sce->refcnt; /* stat_cache_entry_free(sce); */
+                sce = NULL;
+            }
+        }
 
-	const int file_ndx = splaytree_djbhash(name->ptr, len);
-	splay_tree *sptree = sc.files = splaytree_splay(sc.files, file_ndx);
+        if (NULL == sce) {
+            sce = stat_cache_entry_init();
+            buffer_copy_string_len(&sce->name, name->ptr, len);
 
-	if (sptree && (sptree->key == file_ndx)) {
-		/* we have seen this file already and
-		 * don't stat() it again in the same second */
+            /* sptree already splayed to file_ndx in stat_cache_get_entry() */
+            splay_tree * const sptree = sc.files;
+            if (NULL != sptree && sptree->key == file_ndx) {
+                if (refresh < 0) { /* hash collision: replace old entry */
+                    stat_cache_entry_free(sptree->data);
+                } /* else prior sce refcnt was > 1 and decremented above */
+                sptree->data = sce;
+            }
+            else
+                sc.files = splaytree_insert_splayed(sptree, file_ndx, sce);
+        }
+        else {
+            buffer_clear(&sce->etag);
+          #if defined(HAVE_XATTR) || defined(HAVE_EXTATTR)
+            buffer_clear(&sce->content_type);
+          #endif
+        }
 
-		sce = sptree->data;
+        sce->st = st; /*(copy prior to calling fam_dir_monitor())*/
 
-		/* check if the name is the same, we might have a collision */
+      #ifdef HAVE_FAM_H
+        if (sc.stat_cache_engine == STAT_CACHE_ENGINE_FAM) {
+            if (sce->fam_dir) --((fam_dir_entry *)sce->fam_dir)->refcnt;
+            sce->fam_dir = fam_dir_monitor(sc.scf, name->ptr, len, &st);
+          #if 0 /*(performed below)*/
+            if (NULL != sce->fam_dir)
+                /*(may have been invalidated by dir change)*/
+                sce->stat_ts = log_monotonic_secs;
+          #endif
+        }
+      #endif
+    }
 
-		if (buffer_is_equal_string(&sce->name, name->ptr, len)) {
-			if (sc.stat_cache_engine == STAT_CACHE_ENGINE_SIMPLE) {
-				if (sce->stat_ts == cur_ts) {
-					if (final_slash && !S_ISDIR(sce->st.st_mode)) {
-						errno = ENOTDIR;
-						return NULL;
-					}
-					return sce;
-				}
-			}
-		      #ifdef HAVE_FAM_H
-			else if (sc.stat_cache_engine == STAT_CACHE_ENGINE_FAM
-				 && sce->fam_dir) { /* entry is in monitored dir */
-				/* re-stat() periodically, even if monitoring for changes
-				 * (due to limitations in stat_cache.c use of FAM)
-				 * (gaps due to not continually monitoring an entire tree) */
-				if (cur_ts - sce->stat_ts < 16) {
-					if (final_slash && !S_ISDIR(sce->st.st_mode)) {
-						errno = ENOTDIR;
-						return NULL;
-					}
-					return sce;
-				}
-			}
-		      #endif
-		} else {
-			/* collision, forget about the entry */
-			sce = NULL;
-		}
-	}
+    sce->stat_ts = log_monotonic_secs;
+    return sce;
+}
 
-	struct stat st;
-	if (-1 == stat(name->ptr, &st)) {
-		return NULL;
-	}
+stat_cache_entry * stat_cache_get_entry(const buffer * const name) {
 
-	if (NULL == sce) {
+    /* consistency: ensure name in cache does not end in '/' unless root "/"
+     * (but use full path given with stat(), even with trailing '/') */
+    uint32_t len = buffer_clen(name);
+    int final_slash = 0;
+    force_assert(0 != len);
+    if (name->ptr[len-1] == '/') { final_slash = 1; if (0 == --len) len = 1; }
+    /* Note: paths are expected to be normalized before calling stat_cache,
+     * e.g. without repeated '/' */
 
-		/* fix broken stat/open for symlinks to reg files with appended slash on freebsd,osx */
-		/* (local fs_win32_stati64UTF8() checks, but repeat since not obvious)*/
-		if (final_slash && S_ISREG(st.st_mode)) {
-			errno = ENOTDIR;
-			return NULL;
-		}
+    /* check if stat cache entry exists, matches name, and is fresh */
+    int file_ndx;
+    stat_cache_entry *sce =
+      stat_cache_sptree_ndx(&sc.files, &file_ndx, name->ptr, len);
+    int refresh = -1;/* -1 stat cache entry does not exist, or hash collision */
+    if (NULL != sce) {
+        /* check if the name is the same; we might have a hash collision */
+        if (buffer_is_equal_string(&sce->name, name->ptr, len)) {
+            const unix_time64_t cur_ts = log_monotonic_secs;
+            refresh = 1; /* 1 stat cache entry exists, but might need refresh */
+            if (sc.stat_cache_engine == STAT_CACHE_ENGINE_SIMPLE)
+                refresh = (sce->stat_ts != cur_ts);      /* 0 if fresh */
+          #ifdef HAVE_FAM_H
+            else if (sc.stat_cache_engine == STAT_CACHE_ENGINE_FAM
+                     && sce->fam_dir) /* entry is in monitored dir */
+                /* re-stat() periodically, even if monitoring for changes
+                 * (due to limitations in stat_cache.c use of FAM)
+                 * (gaps due to not continually monitoring an entire tree) */
+                refresh = !(cur_ts - sce->stat_ts < 16); /* 0 if fresh */
+          #endif
+        }
+        else /* hash collision; forget about entry */
+            sce = NULL;
+    }
 
-		sce = stat_cache_entry_init();
-		buffer_copy_string_len(&sce->name, name->ptr, len);
+    if (refresh) {
+        sce = stat_cache_refresh_entry(name, len, sce, file_ndx, refresh);
+        if (NULL == sce) return NULL;
+    }
 
-		/* already splayed file_ndx */
-		if (NULL != sptree && sptree->key == file_ndx) {
-			/* hash collision: replace old entry */
-			stat_cache_entry_free(sptree->data);
-			sptree->data = sce;
-		} else {
-			/*sptree =*/ sc.files = splaytree_insert(sptree, file_ndx, sce);
-		}
+    /* fix broken stat/open for symlinks to reg files with appended slash on
+     * old freebsd, osx; fixed in freebsd around 2009:
+     *   https://bugs.freebsd.org/bugzilla/show_bug.cgi?id=21768 */
+    /* (local fs_win32_stati64UTF8() checks, but repeat since not obvious) */
+    if (final_slash && !S_ISDIR(sce->st.st_mode)) {
+        errno = ENOTDIR;
+        return NULL;
+    }
 
-	} else {
-
-		buffer_clear(&sce->etag);
-	      #if defined(HAVE_XATTR) || defined(HAVE_EXTATTR)
-		buffer_clear(&sce->content_type);
-	      #endif
-
-		/* close fd if file changed */
-		if (sce->fd >= 0 && !stat_cache_stat_eq(&sce->st, &st)) {
-			if (1 == sce->refcnt) {
-				close(sce->fd);
-				sce->fd = -1;
-			}
-			else {
-				--sce->refcnt; /* stat_cache_entry_free(sce); */
-				sptree->data = sce = stat_cache_entry_init();
-				buffer_copy_string_len(&sce->name, name->ptr, len);
-			}
-		}
-	}
-
-	sce->st = st; /*(copy prior to calling fam_dir_monitor())*/
-
-#ifdef HAVE_FAM_H
-	if (sc.stat_cache_engine == STAT_CACHE_ENGINE_FAM) {
-		if (sce->fam_dir) --((fam_dir_entry *)sce->fam_dir)->refcnt;
-		sce->fam_dir =
-		  fam_dir_monitor(sc.scf, name->ptr, len, &st);
-	      #if 0 /*(performed below)*/
-		if (NULL != sce->fam_dir) {
-			/*(may have been invalidated by dir change)*/
-			sce->stat_ts = cur_ts;
-		}
-	      #endif
-	}
-#endif
-
-	sce->stat_ts = cur_ts;
-	return sce;
+    return sce;
 }
 
 stat_cache_entry * stat_cache_get_entry_open(const buffer * const name, const int symlinks) {
@@ -1504,12 +1495,8 @@ static void stat_cache_periodic_cleanup(const time_t max_age, const unix_time64_
         max_ndx = 0;
         stat_cache_tag_old_entries(sptree, keys, &max_ndx, max_age, cur_ts);
         for (i = 0; i < max_ndx; ++i) {
-            int ndx = keys[i];
-            sptree = splaytree_splay(sptree, ndx);
-            if (sptree && sptree->key == ndx) {
-                stat_cache_entry_free(sptree->data);
-                sptree = splaytree_delete(sptree, ndx);
-            }
+            sptree = splaytree_splay_nonnull(sptree, keys[i]);
+            sptree = stat_cache_sptree_node_free(sptree);
         }
     } while (max_ndx == sizeof(keys)/sizeof(int));
     sc.files = sptree;
